
from twisted.internet.protocol import DatagramProtocol

from ...share.utils import safe_call, ip_str_to_bin
from ...share.log import logger
from ..task_container import TaskContainer
from ...share.task import Task
from ...share import constants


class ClientConnectionFactory(DatagramProtocol):

    def __init__(self, proxy):
        self.proxy = proxy

    def datagramReceived(self, data, address):
        # address: type: tuple, format: ('127.0.0.1', 62043)

        while data:
            box = self.proxy.app.box_class()
            ret = box.unpack(data)

            if ret == 0:
                # 说明要继续收
                logger.error('buffer incomplete. proxy: %s, ret: %d, read_buffer: %r',
                             self.proxy, ret, data)
                return
            elif ret > 0:
                # 收好了
                # 不能使用双下划线，会导致别的地方取的时候变为 _Gateway__raw_data，很奇怪
                box._raw_data = data[:ret]
                data = data[ret:]
                safe_call(self._on_read_complete, box, address)
                continue
            else:
                # 数据已经混乱了，全部丢弃
                logger.error('buffer invalid. proxy: %s, ret: %d, read_buffer: %r',
                             self.proxy, ret, data)
                data = b''
                return

    def write(self, data, address):
        if self.transport:
            # 要求连接存在
            self.transport.write(data, address)
            self.proxy.stat_counter.client_rsp += 1

            return True

        return False

    def _on_read_complete(self, box, address):
        self.proxy.stat_counter.client_req += 1

        # 获取映射的group_id
        group_id = self.proxy.app.config['GROUP_ROUTER'](box)
        if group_id not in self.proxy.app.group_id_set:
            logger.error('invalid group_id. group_id: %s', group_id)
            return

        client_ip_num, ipv6 = ip_str_to_bin(address[0])

        # 打包成内部通信的task
        task = Task(dict(
            cmd=constants.CMD_WORKER_TASK_ASSIGN,
            client_ip_num=client_ip_num,
            ipv6=ipv6,
            body=box._raw_data,
        ))

        conn = ClientConnection(self, address)

        task_container = TaskContainer(task, conn)
        self.proxy.task_dispatcher.add_task(group_id, task_container)


class ClientConnection(object):

    factory = None
    address = None

    def __init__(self, factory, address):
        self.factory = factory
        self.address = address

    def write(self, data):
        return self.factory.write(data, self.address)




import functools

from events import Events

from ..share.utils import safe_func


class RoutesMixin(object):
    """
    专门做路由管理
    """

    rule_map = None

    def __init__(self):
        self.rule_map = dict()

    def add_route_rule(self, cmd, view_func, endpoint=None, **kwargs):
        old_rule = self.rule_map.get(cmd)
        if old_rule and view_func != old_rule['view_func']:
            raise Exception(
                'duplicate view_func for cmd: {cmd}'.format(cmd=cmd)
            )

        rule = dict(
            endpoint=endpoint or view_func.__name__,
            view_func=view_func,
        )
        rule.update(kwargs)

        self.rule_map[cmd] = rule

    def route(self, cmd, endpoint=None, **kwargs):
        def decorator(func):
            self.add_route_rule(cmd, func, endpoint, **kwargs)
            return func
        return decorator

    def get_route_rule(self, cmd):
        return self.rule_map.get(cmd)


def _reg_event_handler(func):
    @functools.wraps(func)
    def func_wrapper(obj, handler):
        event = getattr(obj.events, func.__name__)
        event += safe_func(handler)

        return handler
    return func_wrapper


class AppEventsMixin(object):
    events = None

    def __init__(self):
        self.events = Events()

    @_reg_event_handler
    def start_worker(self, f):
        """
        创建worker
        f(worker)
        """

    @_reg_event_handler
    def open_conn(self, f):
        """
        worker连接建立后
        f(conn)
        """

    @_reg_event_handler
    def before_first_request(self, f):
        """
        第一个请求，请求解析成功后
        f(request)
        """

    @_reg_event_handler
    def before_request(self, f):
        """
        请求解析成功后
        f(request)
        """

    @_reg_event_handler
    def after_request(self, f):
        """
        执行完route对应的view_func后
        f(request, exc)
        """

    @_reg_event_handler
    def before_response(self, f):
        """
        在 stream.write 之前，传入encode之后的data
        f(conn, response)
        """

    @_reg_event_handler
    def after_response(self, f):
        """
        在 stream.write 之后，传入encode之后的data
        f(conn, response, result)
        """

    @_reg_event_handler
    def close_conn(self, f):
        """
        worker连接close之后
        f(conn)
        """

    @_reg_event_handler
    def stop_worker(self, f):
        """
        停止worker
        f(worker)
        """


class BlueprintEventsMixin(object):

    events = None

    def __init__(self):
        self.events = Events()

    @_reg_event_handler
    def before_request(self, f):
        """
        请求解析成功后
        f(request)
        """

    @_reg_event_handler
    def after_request(self, f):
        """
        执行完route对应的view_func后
        f(request, exc)
        """

    @_reg_event_handler
    def start_app_worker(self, f):
        """
        创建worker
        f(worker)
        """

    @_reg_event_handler
    def open_app_conn(self, f):
        """
        worker连接建立后
        f(conn)
        """

    @_reg_event_handler
    def before_app_first_request(self, f):
        """
        第一次请求，请求解析成功后
        f(request)
        """

    @_reg_event_handler
    def before_app_request(self, f):
        """
        请求解析成功后
        f(request)
        """

    @_reg_event_handler
    def after_app_request(self, f):
        """
        执行完route对应的view_func后
        f(request, exc)
        """

    @_reg_event_handler
    def before_app_response(self, f):
        """
        在 stream.write 之前，传入encode之后的data
        f(conn, response)
        """

    @_reg_event_handler
    def after_app_response(self, f):
        """
        在 stream.write 之后，传入encode之后的data
        f(conn, response, result)
        """

    @_reg_event_handler
    def close_app_conn(self, f):
        """
        worker连接close之后
        f(conn)
        """

    @_reg_event_handler
    def stop_app_worker(self, f):
        """
        停止worker
        f(worker)
        """




import signal
import setproctitle
import os

from .connection import Connection
from ..share.log import logger
from ..share import constants
from ..share.utils import safe_call


class Worker(object):

    type = constants.PROC_TYPE_WORKER

    connection_class = Connection

    group_id = None

    # 是否有效(父进程中代表程序有效，子进程中代表worker是否有效)
    enabled = True

    conn = None

    got_first_request = False

    def __init__(self, app, group_id):
        """
        构造函数
        :return:
        """
        self.app = app
        self.group_id = group_id

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(
            '%s:%s' % (self.type, self.group_id)
        ))

        self._handle_proc_signals()
        self._on_start()

        try:
            address = os.path.join(
                self.app.config['IPC_ADDRESS_DIRECTORY'],
                self.app.config['WORKER_ADDRESS_TPL'] % self.group_id
            )
            self.conn = self.connection_class(self, address, self.app.config['WORKER_CONN_TIMEOUT'])
            self.conn.run()
        except:
            logger.error('exc occur. worker: %s', self, exc_info=True)
        finally:
            self._on_stop()

    def _on_start(self):
        self.app.events.start_worker(self)
        for bp in self.app.blueprints:
            bp.events.start_app_worker(self)

    def _on_stop(self):
        for bp in self.app.blueprints:
            bp.events.stop_app_worker(self)
        self.app.events.stop_worker(self)

    def _handle_proc_signals(self):
        def safe_stop_handler(signum, frame):
            self.enabled = False

            # 关闭读，从而快速退出
            safe_call(lambda: self.conn.shutdown(0))

        # 安全停止
        signal.signal(signal.SIGTERM, safe_stop_handler)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

    def __repr__(self):
        return '<%s name: %s, group_id: %r>' % (
            type(self).__name__, self.app.name, self.group_id
        )



import _thread
import multiprocessing
import signal
import time
import setproctitle

from ..share import constants


class Loader:
    type = constants.PROC_TYPE_LOADER

    app = None

    worker_processes = None

    enabled = True

    def __init__(self, app):
        self.app = app
        self.worker_processes = list()

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(self.type))
        self._handle_signals()

        self._spawn_workers()

        self._monitor_workers()

    def _create_worker(self, group_id):
        process = multiprocessing.Process(
            target=self.app.worker_class(self.app, group_id).run,
        )
        # master不可以独自退出
        process.daemon = False
        process.init_params = dict(
            group_id=group_id,
        )

        process.start()

        return process

    def _spawn_workers(self):
        """
        监控进程
        :return:
        """
        for group in self.app.group_list:
            group_id = group['id']
            for it in range(group['count']):
                self.worker_processes.append(self._create_worker(
                    group_id,
                ))

    def _monitor_workers(self):
        while True:
            for idx, process in enumerate(self.worker_processes):
                if process and not process.is_alive():
                    self.worker_processes[idx] = None

                    if self.enabled:
                        # 需要重启启动process
                        self.worker_processes[idx] = self._create_worker(**process.init_params)

            if not self.enabled and not any(self.worker_processes):
                # 可以彻底停掉了
                break

            time.sleep(0.1)

    def _stop_all(self):
        """
        等所有子进程结束，父进程也退出
        """

        def kill_processes_later(processes, timeout):
            """
            等待一段时间后杀死所有进程
            :param processes:
            :param timeout:
            :return:
            """

            def _kill_processes():
                # 等待一段时间
                time.sleep(timeout)

                for p in processes:
                    if p and p.is_alive():
                        # 说明进程还活着
                        p.kill()

            _thread.start_new_thread(_kill_processes, ())

        self.enabled = False

        # 一定要这样，否则后面kill的时候可能会kill错
        processes = self.worker_processes[:]

        # 如果是终端直接CTRL-C，子进程自然会在父进程之后收到INT信号，不需要再写代码发送
        # 如果直接kill -INT $parent_pid，子进程不会自动收到INT
        # 所以这里可能会导致重复发送的问题，重复发送会导致一些子进程异常，所以在子进程内部有做重复处理判断。
        for p in processes:
            if p:
                p.terminate()

        if self.app.config['STOP_TIMEOUT'] is not None:
            kill_processes_later(processes, self.app.config['STOP_TIMEOUT'])

    def _handle_signals(self):

        def stop_handler(signum, frame):
            """
            等所有子进程结束，父进程也退出
            """
            self._stop_all()

        # TERM为安全结束
        signal.signal(signal.SIGTERM, stop_handler)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        # 热重启
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

    def __repr__(self):
        return '<%s name: %s>' % (
            type(self).__name__, self.app.name
        )




import sys
import os
import json
from collections import Counter

from .share.log import logger
from .share.mixins import RoutesMixin, AppEventsMixin
from .share import constants
from .share.config import ConfigAttribute, Config
from .share.utils import import_module_or_string


class Burst(RoutesMixin, AppEventsMixin):

    # 配置都放到 burst 里，而和proxy或者worker直接相关的类，则放到自己的部分
    ############################## configurable begin ##############################
    
    name = ConfigAttribute('NAME')

    box_class = ConfigAttribute('BOX_CLASS',
                                get_converter=import_module_or_string)
    master_class = ConfigAttribute('MASTER_CLASS',
                                   get_converter=import_module_or_string)
    proxy_class = ConfigAttribute('PROXY_CLASS',
                                  get_converter=import_module_or_string)
    loader_class = ConfigAttribute('LOADER_CLASS',
                                   get_converter=import_module_or_string)
    worker_class = ConfigAttribute('WORKER_CLASS',
                                   get_converter=import_module_or_string)
    request_class = ConfigAttribute('REQUEST_CLASS',
                                    get_converter=import_module_or_string)

    group_list = ConfigAttribute('GROUP_LIST')

    group_id_set = None

    debug = ConfigAttribute('DEBUG')

    ############################## configurable end   ##############################

    config = None
    blueprints = None

    def __init__(self):
        """
        构造函数
        :return:
        """
        RoutesMixin.__init__(self)
        AppEventsMixin.__init__(self)

        self.config = Config(defaults=constants.DEFAULT_CONFIG)
        self.blueprints = list()

    def register_blueprint(self, blueprint):
        blueprint.register_to_app(self)

    def run(self, host=None, port=None, debug=None):
        self._merge_routes()

        if host is not None:
            self.config.update({
                'HOST': host,
            })

        if port is not None:
            self.config.update({
                'PORT': port,
            })

        if debug is not None:
            self.debug = debug

        # run之前才生成
        self.group_id_set = set([group['id'] for group in self.group_list])

        # 只要没有这个环境变量，就是主进程
        str_burst_env = os.getenv(self.config['CHILD_PROCESS_ENV_KEY'])

        if not str_burst_env:
            # 主进程
            logger.info('Running server. name: %s, address: %s, admin_address: %s, debug: %s',
                        self.name, (self.config['HOST'], self.config['PORT']), self.config['ADMIN_ADDRESS'],
                        self.debug)
            self.master_class(self).run()
        else:
            burst_env = json.loads(str_burst_env)
            if burst_env['type'] == constants.PROC_TYPE_PROXY:
                # proxy
                self.proxy_class(self).run()
            elif burst_env['type'] == constants.PROC_TYPE_LOADER:
                # loader
                self.loader_class(self).run()
            else:
                logger.error('unknown proc type. env: %s', burst_env)
                return

    def make_proc_name(self, subtitle):
        """
        获取进程名称
        :param subtitle:
        :return:
        """
        proc_name = '[%s:%s %s] %s' % (
            constants.NAME,
            subtitle,
            self.name,
            ' '.join([sys.executable] + sys.argv)
        )

        return proc_name

    def change_group_config(self, group_id, count):
        """
        修改组配置，master和proxy都会用到
        :param group_id:
        :param count:
        :return:
        """

        # 不能设置成个奇怪的值就麻烦了
        if not isinstance(group_id, int) or not isinstance(count, int):
            return False

        for group in self.group_list:
            if group_id == group['id']:
                group['count'] = count
                break
        else:
            self.group_list.append(dict(
                id=group_id,
                count=count
            ))

        return True

    def _merge_routes(self):
        """
        合并routes
        :return:
        """

        for bp in self.blueprints:
            for cmd, rule in bp.rule_map.items():
                new_rule = dict(rule)
                new_rule['blueprint'] = bp
                new_rule['endpoint'] = '.'.join([bp.name, new_rule['endpoint']])

                self.add_route_rule(cmd, **new_rule)

    def __repr__(self):
        return '<%s name: %s>' % (
            type(self).__name__, self.name
        )




import os
import copy
import json
import sys
import subprocess
import time
import signal
import _thread
import setproctitle
from netkit.box import Box
from netkit.contrib.tcp_client import TcpClient

from ..share.log import logger
from ..share.utils import safe_call
from ..share import constants


class Master(object):
    """
    master相关
    """

    type = constants.PROC_TYPE_MASTER

    app = None

    # 是否有效
    enabled = True

    # 子进程们
    child_processes = None

    def __init__(self, app):
        """
        构造函数
        :return:
        """
        self.app = app
        self.child_processes = dict()

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(self.type))

        self._handle_proc_signals()

        proxy = self._spawn_proxy()
        if not proxy:
            # 启动proxy失败，就应该直接返回
            self._stop_all()
            return

        self.child_processes[constants.PROC_TYPE_PROXY] = proxy

        # 等待proxy启动，为了防止worker在连接的时候一直报connect失败的错误
        if not self._wait_proxy():
            # 有可能ctrl-c终止，这个时候就要直接返回了
            self._stop_all()
            return

        loader = self._spawn_loader()
        if not loader:
            # 启动loader失败，就应该直接返回
            self._stop_all()
            return

        self.child_processes[constants.PROC_TYPE_LOADER] = loader

        _thread.start_new_thread(self._connect_to_proxy, ())

        self._monitor_child_processes()

    def _wait_proxy(self):
        """
        尝试连接proxy，如果连接成功，说明proxy启动起来了
        :return:
        """
        address = os.path.join(
            self.app.config['IPC_ADDRESS_DIRECTORY'],
            self.app.config['MASTER_ADDRESS']
        )
        client = TcpClient(Box, address=address)

        while self.enabled:
            try:
                client.connect()
                # 连接成功后，就关闭连接
                client.close()
                return True
            except:
                time.sleep(0.1)
                continue

        return False

    def _connect_to_proxy(self):
        """
        连接到proxy，因为有些命令要发过来
        :return:
        """
        address = os.path.join(
            self.app.config['IPC_ADDRESS_DIRECTORY'],
            self.app.config['MASTER_ADDRESS']
        )
        client = TcpClient(Box, address=address)

        while self.enabled:
            try:
                if client.closed():
                    client.connect()
            except:
                # 只要连接失败
                logger.error('connect fail. master: %s, address: %s', self, address)
                time.sleep(1)
                continue

            # 读取的数据
            box = client.read()
            if not box:
                logger.info('connection closed. master: %s', self)
                continue

            logger.info('box received. master: %s, box: %s', self, box)

            safe_call(self._handle_proxy_data, box)

    def _handle_proxy_data(self, box):
        """
        处理从proxy过来的box
        :param box:
        :return:
        """

        if box.cmd == constants.CMD_ADMIN_RELOAD:
            self._reload_workers()
        elif box.cmd == constants.CMD_ADMIN_STOP:
            self._stop_all()

    def _start_child_process(self, proc_env):
        worker_env = copy.deepcopy(os.environ)
        worker_env.update({
            self.app.config['CHILD_PROCESS_ENV_KEY']: json.dumps(proc_env)
        })

        args = [sys.executable] + sys.argv
        try:
            inner_p = subprocess.Popen(args, env=worker_env)
            inner_p.proc_env = proc_env
            return inner_p
        except:
            logger.error('exc occur. app: %s, args: %s, env: %s',
                         self, args, worker_env, exc_info=True)
            return None

    def _spawn_proxy(self):
        proc_env = dict(
            type=constants.PROC_TYPE_PROXY
        )
        return self._start_child_process(proc_env)

    def _spawn_loader(self):
        proc_env = dict(
            type=constants.PROC_TYPE_LOADER
        )
        return self._start_child_process(proc_env)

    def _monitor_child_processes(self):
        while 1:
            for proc_type, p in self.child_processes.items():

                if p and p.poll() is not None:
                    # 说明退出了
                    proc_env = p.proc_env
                    self.child_processes[proc_type] = None

                    if self.enabled:
                        # 如果还要继续服务
                        p = self._start_child_process(proc_env)
                        self.child_processes[proc_type] = p

            if not any(self.child_processes.values()):
                # 没活着的proxy和worker了
                break

            # 时间短点，退出的快一些
            time.sleep(0.1)

    def _kill_processes_later(self, processes, timeout):
        """
        等待一段时间后杀死所有进程
        :param processes:
        :param timeout:
        :return:
        """
        def _kill_processes():
            # 等待一段时间
            time.sleep(timeout)

            for p in processes:
                if p and p.poll() is None:
                    # 说明进程还活着
                    safe_call(p.send_signal, signal.SIGKILL)

        _thread.start_new_thread(_kill_processes, ())

    def _stop_all(self):
        """
        停止所有进程
        :return:
        """
        self.enabled = False

        processes = list(self.child_processes.values())

        for p in processes:
            if p:
                safe_call(p.send_signal, signal.SIGTERM)

        # 不能使用强制kill的方式，因为loader即使被强制kill，如果worker还在正常运行的话，问题更大

        # if self.app.config['STOP_TIMEOUT'] is not None:
        #    self._kill_processes_later(processes, self.app.config['STOP_TIMEOUT'])

    def _reload_workers(self):
        """
        reload workers
        :return:
        """

        loader = self.child_processes.get(constants.PROC_TYPE_LOADER)
        if loader:
            safe_call(loader.send_signal, signal.SIGTERM)

    def _handle_proc_signals(self):
        def stop_handler(signum, frame):
            """
            等所有子进程结束，父进程也退出
            """
            self._stop_all()

        def reload_handler(signum, frame):
            """
            让所有子进程重新加载
            """
            self._reload_workers()

        # INT为强制结束
        signal.signal(signal.SIGINT, stop_handler)
        # TERM为安全结束
        signal.signal(signal.SIGTERM, stop_handler)
        # HUP为热更新
        signal.signal(signal.SIGHUP, reload_handler)

    def __repr__(self):
        return '<%s name: %s>' % (
            type(self).__name__, self.app.name
        )



from setuptools import setup, find_packages

setup(
    name='burst',
    version='3.1.37',
    zip_safe=False,
    platforms='any',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    python_requires='>=3',
    scripts=['burst/bin/burstctl'],
    install_requires=['setproctitle', 'twisted', 'events', 'netkit', 'click'],
    url='https://github.com/dantezhu/burst',
    license='MIT',
    author='dantezhu',
    author_email='zny2008@gmail.com',
    description='twisted with master, proxy and worker',
)
